from numpy import where, array, mean, zeros
import torch
from transformers import AutoTokenizer, AutoModel
import logging
from offline_results.common.config import ZERO_VECTOR_DIMENSION
logging.basicConfig(level=logging.INFO)

class FeatureVectors:

    def __init__(
            self,
            tokenizer: str,
            model: str,
            layers=None
    ):
        """
        Constructor to accept AutoModel Tokenizer and Model.
        If required the layers can be specified else,
        the default value shall be user
        @param tokenizer: Transformers Automodel Tokenizer
        @param model: Transformers Automodel Model
        @param layers: Number of language model layers
        to use for generating embeddings
        """
        self.tokenizer = AutoTokenizer.from_pretrained(
            tokenizer
        )
        self.model = AutoModel.from_pretrained(
            pretrained_model_name_or_path=model,
            output_hidden_states=True
        )
        # Use last four layers by default
        self.layers = [-4, -3, -2, -1] \
            if layers is None else layers

    def get_word_idx(
            self,
            sent: str,
            word: str
    ):
        """
        Get index of a specified word in a
        phrase/sentence
        @param sent: phrase / sentence
        @param word: word within the
        phrase / sentence
        @return: index of the word in the phrase
        """
        return sent.split(" ").index(word)

    def get_hidden_states(
            self,
            encoded,
            token_ids_word,
            model,
            layers
    ):
        """
        Push input IDs through model. Stack and sum
        `layers` (last four by default). Select only
        those subword token outputs that belong to
        our word of interest and average them.
        @param encoded: encoded tensor format
        @param token_ids_word: subword token indices
        for the considered word
        @param model: Automodel model
        @param layers: list of layers to be considered
        @return: numpy array of feature embeddings
        """
        with torch.no_grad():
            output = model(**encoded)

        # Get all hidden states
        states = output.hidden_states

        # Stack and sum all requested layers
        output = torch.stack(
            [states[i] for i in layers]
        ).sum(0).squeeze()

        # Only select the tokens that constitute
        # the requested word
        word_tokens_output = output[token_ids_word]

        return word_tokens_output.mean(dim=0)

    def get_word_vector(
            self,
            sent,
            idx,
            tokenizer,
            model,
            layers
    ):
        """
        Get a word vector by first tokenizing the input
        sentence, getting all token idxs that make up the
        word of interest, and then `get_hidden_states`.
        @param sent: phrase / sentence
        @param idx: index of the word in phrase / sentence
        @param tokenizer: Automodel Tokenizer
        @param model: Automodel Model
        @param layers: list of layers to be used from the
         pre-trained language model
        @return: Word embeddings
        """
        encoded = tokenizer.encode_plus(
            sent,
            return_tensors="pt"
        )
        # get all token indices that belong
        # to the word of interest
        token_ids_word = where(
            array(encoded.word_ids()) == idx
        )

        return self.get_hidden_states(
            encoded,
            token_ids_word,
            model,
            layers
        )

    def controller(
            self,
            phrases: list
    ):
        """
        Driver function to generate feature representation
        for a list of words / phrases
        @param phrases: list of words / phrases
        @return: Word embeddings for the passed set
        of phrases / words
        """
        phrases_vectors = []
        if len(phrases) == 0:
            return zeros((ZERO_VECTOR_DIMENSION,),
                         dtype=float)

        for index, phrase in enumerate(phrases):

            logging.info(" Generating embeddings for phrase "
                         + phrase + " ( " + str(index+1) +
                         " of " + str(len(phrases)) + " )")

            word_vectors = []
            words = phrase.split(" ")
            for word in words:
                logging.info(" ----------- For word" + word)
                idx = self.get_word_idx(
                    sent=phrase,
                    word=word
                )

                vector = self.get_word_vector(
                    sent=phrase,
                    idx=idx,
                    tokenizer=self.tokenizer,
                    model=self.model,
                    layers=self.layers
                )
                logging.info("Length of vector generated = " +
                             str(len(vector)))

                word_vectors.append(array(vector.tolist()))
            phrases_vectors.append(mean(word_vectors, axis=0))

        return mean(phrases_vectors, axis=0)
