from offline_results.common.config import RATING_VALUES, TOKENIZER, MODEL, DISCARD_ATTRIBUTES
from offline_results.common.constants import CSV_EXTENSION, RATING, \
    CATEGORY, CUSTOMER_ID, SUBCATEGORY, TAGS, CATEGORY_ID, \
    CATEGORY_EN, SUBCATEGORY_ID, SUBCATEGORY_EN, TAGS_ID, TAGS_NAME, \
    SOLO_FEATURE_LIST, VIDEO_ID1, CONTENT_ID
from offline_results.recommendation.utils import RecommendationUtils
from offline_results.common.read_write_from_s3 import ConnectS3
from offline_results.recommendation.homepage_cluster_category.get_feature_vectors import FeatureVectors
from pandas import DataFrame, merge
from graphdb.graph import GraphDb
from functools import reduce
import logging
logging.basicConfig(level=logging.INFO)

class PrepareData:

    def __init__(
            self,
            connection_object,
            resource: object,
            bucket_name: str,
            object_name: str,
            keep_rating: bool = False,
            keep_category: bool = False,
            keep_subcategory: bool = False,
            keep_tags: bool = False,
            vectorization_tokenizer: str = TOKENIZER,
            vectorization_model=MODEL
    ):
        """
        Constructor to create instance member properties
        for s3 communication. The instance member properties
        also include boolean indicators of the features that
        shall be used subsequently by the recommendation model.
        @param connection_object: graphDB connection object
        @param resource: S3 resource object
        @param bucket_name: S3 bucket name
        @param object_name: S3 object name
        @param keep_rating: Boolean indicator for using rating
        @param keep_category: Boolean indicator for using category
        @param keep_subcategory: Boolean indicator for using subcategory
        @param keep_tags: Boolean indicator for using tags
        @param vectorization_tokenizer: Tokenizer used to generate
        language model embeddings
        @param vectorization_model: Model to generate embeddings
        """
        self.graph = GraphDb.from_connection(connection_object)
        self.resource = resource
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.keep_rating = keep_rating
        self.keep_category = keep_category
        self.keep_subcategory = keep_subcategory
        self.keep_tags = keep_tags
        self.feature_vectors = FeatureVectors(
            tokenizer=vectorization_tokenizer,
            model=vectorization_model
        )

    def fetch_data(
            self,
            attribute: str
    ):
        """
        To retrieve preference information for a specified attribute.
        This information can either be retrieved from S3 or
        alternatively, directly from the graphDB
        @param attribute: the preference attribute for which the
        information needs to be retrieved.
        @return: Attribute preference dataframe object pandas
        """

        return ConnectS3().read_csv_from_s3(
            bucket_name=self.bucket_name,
            object_name=self.object_name +
            attribute + CSV_EXTENSION,
            resource=self.resource
        )

    def format_ubd(
            self,
            ubd
    ):
        """
        Basic formatting of attributes in UBD
        @param ubd: Dataframe object pandas
        @return: dataframe object pandas
        """
        ubd = ubd[[CUSTOMER_ID, VIDEO_ID1]]
        ubd[CUSTOMER_ID] = ubd[CUSTOMER_ID].astype(str)
        ubd = ubd.rename(columns={VIDEO_ID1: CONTENT_ID})
        return ubd

    def fetch_and_format_ubd(
            self
    ) -> DataFrame:
        """
        Fetches UBD video_measure.csv data from S3,
        filters out unwanted attributes and returns
        homepage id for all the content listed in
        UBD dataframe object pandas
        @return: Dataframe object pandas
        """

        ubd = self.fetch_data(attribute="video_measure")
        ubd = self.format_ubd(ubd)

        content_homepage_mapping = DataFrame(
            RecommendationUtils.get_homepage_for_contents(
                graph=self.graph))

        return merge(ubd, content_homepage_mapping,
                     on=CONTENT_ID, how="inner").reset_index(drop=True)

    def format_preferences(
            self,
            data: DataFrame,
            data_key: str,
            attribute: str,
            solo_feature: bool = False
    ) -> DataFrame:
        """
        For the preference information related to the specified
        attribute, format the layout of the dataframe object by
        replacing 1's with corresponding preference value string.
        Since for the solo features such as rating and attribute1,
        there is no particular identifier, their values need not
        be type casted. This condition is captured using the solo_
        feature flag.
        @param data: Dataframe object pandas
        @param data_key: The key attribute in data variable. This
        attribute used to generate the final result returned
        by the function.
        @param attribute: String value for the preference attribute.
        @param solo_feature: Boolean indicator. When preparing the
        result, if solo_feature is set to True, the values are not
        type casted to corresponding integers.
        @return: Formatted dataframe object pandas with preferences
        within a single dataframe column.
        """
        formatted_preferences = DataFrame()
        formatted_preferences[data_key] = data[data_key]
        data = data.drop(columns=[data_key])

        # Since the key column has been dropped from the
        # 'data' variable, all the columns currently in this
        # dataframe are preference attributes. Thus, iterating
        # through each attribute to reformat the preferences.

        for feature in data.columns:
            logging.info("Formatting the feature " + feature)
            feature_val = feature.split("_")[-1]

            if not solo_feature and \
                    feature_val not in DISCARD_ATTRIBUTES:
                feature_val = int(feature_val)

            data[feature] = [feature_val
                             if val == 1 else 0
                             for val in data[feature]]

        # Merging all the formatted attribute preferences under a
        # single dataframe column in the result

        formatted_preferences[attribute] = \
            data[data.columns].apply(list, axis=1)

        # Remove all the preference values that are either 0 or nan.
        # These values do not contribute to the preferences of
        # a given user

        formatted_preferences[attribute] = \
            [list(filter(lambda val: val not in DISCARD_ATTRIBUTES,
                         preferences))
             for preferences in formatted_preferences[attribute]]

        return formatted_preferences

    def map_static_preferences(
            self,
            data: DataFrame,
            preference_map: DataFrame,
            map_key: str,
            map_val: str,
            attribute: str
    ):
        """
        To map preference attributes whose static nodes
        have been dumped into graphDB. The function
        queries the graphDB to retrieve all nodes of
        specified attribute node label and returns the
        key-value pair for the requested property.
        @param data: Dataframe object pandas
        @param preference_map: Dataframe object pandas
        @param map_key: The attribute to be considered
        as the key in preference_map
        @param map_val: The attribute to be considered
        as the value map to the key in the preference_map
        @param attribute: Attribute value to be considered
        @return: Dataframe object pandas mapped for the
        specified attribute
        """
        record_count = len(data)

        for index in range(record_count):

            logging.info("Working on " + attribute
                         + " record " + str(index))

            preferences = data.loc[index, attribute]

            preferences = preference_map[
                preference_map[map_key].isin(preferences)
            ].reset_index(drop=True)

            data.at[index, attribute] = \
                preferences[map_val].tolist()

        return data

    def map_solo_feature_preferences(
            self,
            data: DataFrame,
            preference_map: list,
            attribute: str
    ) -> DataFrame:
        """
        To map preference attributes whose static nodes
        have NOT been dumped into graphDB. The function
        uses the pre-defined config mappings to retrieve
        all possible values of the specified attribute
        and returns the vectorised form for the
        requested property.
        @param data: Dataframe object pandas
        @param preference_map: list of possible values
        for the specified attribute
        @param attribute: Attribute value to be considered
        @return:  Dataframe object pandas vectorized for the
        specified attribute
        """
        for index in range(len(data)):

            logging.info("Working on " + attribute
                         + " record " + str(index))

            feature_preference = data.loc[index, attribute]

            data.at[index, attribute] = \
                [1 if preference.upper() in feature_preference
                 else 0 for preference in preference_map]

        return data

    def get_preferences(
            self,
            attribute: str,
            label: str = None,
            map_key: str = None,
            map_val: str = None,
            solo_feature: bool = False,
    ) -> DataFrame:
        """
        Fetch and format the preferences for a given attribute.
        @param attribute: Value string for an attribute
        @param label: Node label to search for in graphDB for
        preparing the preference key-value mappings
        @param map_key: Key node property to look for in graphDB
        @param map_val: Value node property to look for in graphDB
        @param solo_feature: Boolean indicator. Used during
        the formatting of the result.
        @return: Dataframe object pandas with formatted
        preference credentials
        """
        logging.info("Fetching preferences data for" + attribute)
        preference_data = self.fetch_data(
            attribute=attribute
        )

        logging.info("Formatting preference attributes....")
        formatted_preferences = self.format_preferences(
            data=preference_data,
            data_key=CUSTOMER_ID,
            attribute=attribute,
            solo_feature=solo_feature
        )

        # if the feature under consideration does not
        # have any independent static nodes in the graphDB,
        # return the result at this stage. The further
        # procedure shall be vectorization of values
        # under this feature.
        if label is None:
            return formatted_preferences

        preference_map = RecommendationUtils.get_attribute_map(
            graph=self.graph,
            node_label=label,
            key_node_property=map_key,
            value_node_property=map_val
        )

        return self.map_static_preferences(
            data=formatted_preferences,
            preference_map=DataFrame(preference_map[0]),
            map_key=map_key,
            map_val=map_val,
            attribute=attribute
        )

    def get_mapped_feature_preferences(
            self
    ):
        """
        Driver function for generating Homepage Cluster
        Category recommendation model input data.
        Solo features are the features which do not have a separate
        collection in the V+ DB. These features include
        rating, attribute1 etc.
        @return: Dataframe object pandas
        """
        preference_attributes = []

        # prepare user-wise rating preferences
        # if rating is to be included
        if self.keep_rating:
            rating_preferences = \
                self.get_preferences(
                    attribute=RATING,
                    solo_feature=True
                )
            preference_attributes.append(
                self.map_solo_feature_preferences(
                    data=rating_preferences,
                    preference_map=RATING_VALUES,
                    attribute=RATING
                )
            )

        # prepare user-wise category preferences
        # if category is to be included
        if self.keep_category:
            preference_attributes.append(
                self.get_preferences(
                    attribute=CATEGORY,
                    solo_feature=False,
                    label=CATEGORY,
                    map_key=CATEGORY_ID,
                    map_val=CATEGORY_EN
                )
            )

        # prepare user-wise subcategory preferences
        # if subcategory is to be included
        if self.keep_subcategory:
            preference_attributes.append(
                self.get_preferences(
                    attribute=SUBCATEGORY,
                    solo_feature=False,
                    label=SUBCATEGORY,
                    map_key=SUBCATEGORY_ID,
                    map_val=SUBCATEGORY_EN
                )
            )

        # prepare user-wise tags preferences
        # if tags is to be included
        if self.keep_tags:
            preference_attributes.append(
                self.get_preferences(
                    attribute=TAGS,
                    solo_feature=False,
                    label=TAGS,
                    map_key=TAGS_ID,
                    map_val=TAGS_NAME
                )
            )

        # Merge results from the above preferences
        # set to True and return merged result
        return reduce(
            lambda l, r: merge(l, r,
                               on=CUSTOMER_ID,
                               how='inner'),
            preference_attributes
        )

    def get_preference_embeddings(
            self,
            feature_preferences: DataFrame,
            data_key: str
    ) -> DataFrame:
        """
        Get feature representation for the preferences
        of a considered attribute
        @param feature_preferences:Dataframe object pandas
        @param data_key: key attribute in the dataframe object
        @return: Dataframe object pandas
        """
        for preference in feature_preferences.columns:

            logging.info("Generating preference " +
                         "embeddings for " + preference)

            if preference == data_key \
                    or preference in SOLO_FEATURE_LIST:
                continue

            feature_preferences[preference] = \
                [self.feature_vectors.controller(phrases=vector)
                 for vector in feature_preferences[preference]]

        return feature_preferences

    def controller(
            self
    ):
        """
        Driver function to retrieve preference data, map the
        preferences to the corresponding values, adn generate
        vector representation using BERT pre-trained
        language model
        @return: Dataframe object pandas
        """
        preferences = self.get_mapped_feature_preferences()

        vectorized_preferences = self.get_preference_embeddings(
            feature_preferences=preferences,
            data_key=CUSTOMER_ID
        )

        vectorized_preferences[CUSTOMER_ID] = \
            vectorized_preferences[CUSTOMER_ID].astype(str)

        ubd = self.fetch_and_format_ubd()

        return merge(vectorized_preferences, ubd,
                     on=CUSTOMER_ID, how="inner")