from datetime import datetime
from itertools import chain

from numpy import flip
from pandas import DataFrame, merge, concat

from offline_results.common.constants import CUSTOMER_ID, CONTENT_ID, VIEW_COUNT, VIEW_HISTORY, IS_PAYTV, \
    NO_PAY_TV_CONTENT, PAY_TV_CONTENT, YES, NO, INDEX, LEFT, TIME_CREATED, HOMEPAGE_ID, RECOMMENDED_CONTENT_ID, \
    RECOMMENDATION_SCORE
from offline_results.recommendation.homepage_recommendation.core.with_cf_item_to_item.query_utils import QueryUtils
from offline_results.similarity.content_profile.similarity_all_contents import SimilarityAllContents
from offline_results.similarity.content_profile.similarity_homepage_contents import SimilarityHomepageContents


class ItemToItemUtils():

    @staticmethod
    def user_content_network_to_dataframe_with_content_paytv_flag(response, content_paytv_flag) -> DataFrame:

        viewed_feature_list = []

        for index, ele in enumerate(response):
            for value in range(len(response[index])):
                customer_id = response[index][value][0][CUSTOMER_ID]
                view_count = response[index][value][1][VIEW_COUNT]
                view_history = response[index][value][1][VIEW_HISTORY]
                content_id = response[index][value][2][CONTENT_ID]
                view_dict = {CUSTOMER_ID: customer_id,
                             CONTENT_ID: content_id,
                             VIEW_COUNT: view_count,
                             VIEW_HISTORY: view_history,
                             }
                viewed_feature_list.append(view_dict)

        result = DataFrame(viewed_feature_list).reset_index(drop=True)

        result[IS_PAYTV] = content_paytv_flag
        return result

    @staticmethod
    def user_content_network_to_dataframe(graph):

        user_content_paytv_df = ItemToItemUtils.user_content_network_to_dataframe_with_content_paytv_flag(
            QueryUtils.query_log_data_for_user_including_paytv_status(graph), YES)
        user_content_no_paytv_df = ItemToItemUtils.user_content_network_to_dataframe_with_content_paytv_flag(
            QueryUtils.query_log_data_for_user_including_no_paytv_status(graph), NO)

        return user_content_paytv_df, user_content_no_paytv_df

    @staticmethod
    def queried_network_response_to_dataframe(response) -> DataFrame:
        return DataFrame(list(chain.from_iterable(response)))

    @staticmethod
    def homepage_network_to_dataframe_with_content_paytv_flag(graph):
        homepage_paytv_df = ItemToItemUtils.queried_network_response_to_dataframe(
            QueryUtils.query_homepage_data_for_paytv_contents(graph))
        homepage_no_paytv_df = ItemToItemUtils.queried_network_response_to_dataframe(
            QueryUtils.query_homepage_data_for_no_paytv_contents(graph))
        return homepage_paytv_df, homepage_no_paytv_df

    @staticmethod
    def similarity_dict_of_dict_to_dataframe(similarity_response) -> DataFrame:
        df = DataFrame.from_dict(similarity_response, orient=INDEX).fillna(0)
        temp = flip(df.values.argsort(), 1)
        new_df = DataFrame(df.columns[temp]).iloc[:, : 10]
        new_df[CONTENT_ID] = df.index
        return new_df

    @staticmethod
    def get_similar_contents_based_on_all(graph):
        paytv_df = ItemToItemUtils.similarity_dict_of_dict_to_dataframe(
            SimilarityAllContents.prepare_similarity_based_on_all_content(PAY_TV_CONTENT, None, graph))
        no_paytv_df = ItemToItemUtils.similarity_dict_of_dict_to_dataframe(
            SimilarityAllContents.prepare_similarity_based_on_all_content(NO_PAY_TV_CONTENT, None, graph))
        return paytv_df, no_paytv_df

    @staticmethod
    def find_maximum_view_count(data) -> DataFrame:
        return data.sort_values(VIEW_COUNT, ascending=False).drop_duplicates([CUSTOMER_ID])

    @staticmethod
    def merge_on_content_id(dataframe1, dataframe2) -> DataFrame:
        return merge(dataframe1, dataframe2, on=[CONTENT_ID], how=LEFT)

    @staticmethod
    def merge_on_homepage_id(dataframe1, dataframe2) -> DataFrame:
        return merge(dataframe1, dataframe2, on=[HOMEPAGE_ID], how=LEFT)

    @staticmethod
    def add_time_created_attribute(dataframe) -> DataFrame:
        dataframe[TIME_CREATED] = datetime.now()
        return dataframe

    @staticmethod
    def map_homepage_and_content_status(graph, paytv_df, no_paytv_df):

        homepage_status_df = ItemToItemUtils.queried_network_response_to_dataframe(
            QueryUtils.query_static_homepage_data_for_homepage_status(graph))

        paytv_content_status_df = ItemToItemUtils.queried_network_response_to_dataframe(
            QueryUtils.query_pay_tv_content_data_for_content_status(graph))

        no_paytv_content_status_df = ItemToItemUtils.queried_network_response_to_dataframe(
            QueryUtils.query_no_pay_tv_content_data_for_content_status(graph))

        homepage_status_df_paytv = ItemToItemUtils.merge_on_homepage_id(paytv_df, homepage_status_df)
        homepage_status_df_no_paytv = ItemToItemUtils.merge_on_homepage_id(no_paytv_df, homepage_status_df)

        paytv_content_status_df = ItemToItemUtils.merge_on_content_id(paytv_df, paytv_content_status_df)
        no_paytv_content_status_df = ItemToItemUtils.merge_on_content_id(no_paytv_df, no_paytv_content_status_df)

        final_df_paytv = merge(homepage_status_df_paytv, paytv_content_status_df)
        final_df_no_paytv = merge(homepage_status_df_no_paytv, no_paytv_content_status_df)

        return final_df_paytv, final_df_no_paytv

    @staticmethod
    def get_similar_contents_based_on_homepage_id(content_label, homepage_id, graph):

        cc_similarity = SimilarityHomepageContents.prepare_similarity_based_on_homepage_id(content_label,
                                                                                           homepage_id, graph)
        df = DataFrame(cc_similarity).T
        df[CONTENT_ID] = df.index
        temp_df = DataFrame([*df[0]], df.index).stack().rename_axis([None, RECOMMENDED_CONTENT_ID]).reset_index(1,
                                                                                                                name=RECOMMENDATION_SCORE)
        recommendation_df = df[[CONTENT_ID]].join(temp_df)
        recommendation_df[HOMEPAGE_ID] = homepage_id

        return recommendation_df

    @staticmethod
    def prepare_user_wise_recommendation_results_based_on_homepage_id(customer_id, content_label, graph, dataframe):
        df = dataframe[dataframe[CUSTOMER_ID] == customer_id]
        final_recommendation_df = DataFrame()
        for idx, val in df.iterrows():
            recommendation_df = ItemToItemUtils.get_similar_contents_based_on_homepage_id(content_label,
                                                                                          val[HOMEPAGE_ID], graph)
            recommendation_df_for_user = merge(df, recommendation_df, on=[HOMEPAGE_ID, CONTENT_ID])
            final_recommendation_df = concat([final_recommendation_df, recommendation_df_for_user], axis=0)
        return ItemToItemUtils.add_time_created_attribute(final_recommendation_df)
