from datetime import datetime
import time
import pandas as pd

from offline_results.common.constants import PAY_TV, NO_PAY_TV, SERVICE_NAME, ALL_CONTENT_BASED, HOMEPAGE_ID, \
    CONTENT_ID, CREATED_ON, SEASON, HOMEPAGE_ID_BASED, VIEW_COUNT, LEFT, \
    DEFAULT_SEASONAL_MODULE_NAME, REC_TYPE, SEASONAL_FALLBACK, RECORDS
from offline_results.recommendation.homepage_recommendation.seasonal.seasonal_model import SeasonalModel
from offline_results.repository.graph_db_connection import ANGraphDb
from offline_results.utils.logger import Logging

graph = ANGraphDb.new_connection_config().graph


class SeasonalModelDefault:

    @staticmethod
    def get_dict_format_output(df, key_prefix, homepage_id_wise):
        if homepage_id_wise:
            output_dict = {}
            unique_homepage_id = df[HOMEPAGE_ID].unique()
            for homepage_id in unique_homepage_id:
                key_prefix_cls_hid = key_prefix + ":" + str(homepage_id)
                homepage_wise_df = df.loc[df[HOMEPAGE_ID] == homepage_id]
                homepage_wise_df = homepage_wise_df[[CONTENT_ID, CREATED_ON, REC_TYPE, SEASON]]
                output_dict[key_prefix_cls_hid] = homepage_wise_df.to_dict(RECORDS)
        else:
            df = df[[CONTENT_ID, HOMEPAGE_ID, CREATED_ON, REC_TYPE, SEASON]]
            output_dict = {key_prefix: df.to_dict(RECORDS)}
        return output_dict

    @staticmethod
    def default_seasonal_algorithm(graph, user_label, homepage_id_wise):
        default_status = True
        history_content_df, content_homepage_id_mapping = SeasonalModel.fetch_historical_data(graph, user_label,
                                                                                              homepage_id_wise)
        season_df, current_season_name = SeasonalModel.get_specific_seasons(history_content_df, default_status)
        Logging.info("Calculating Seasonal Score from " + user_label + " UBD data - DEFAULT MODE")
        season_df_view_count = season_df.groupby([CONTENT_ID]).size().to_frame(VIEW_COUNT)
        Logging.info("Preparing Seasonal Model Dataframe Output - DEFAULT MODE")
        final_recommendation_df = season_df_view_count.sort_values(by=VIEW_COUNT, ascending=False)
        final_recommendation_df[CREATED_ON] = datetime.utcnow().isoformat()
        seasonal_content_df = final_recommendation_df[[VIEW_COUNT, CREATED_ON]].rename_axis(CONTENT_ID).reset_index()
        seasonal_content_df[SEASON] = current_season_name
        seasonal_content_df[REC_TYPE] = SEASONAL_FALLBACK

        if homepage_id_wise:
            seasonal_content_df = pd.merge(seasonal_content_df, content_homepage_id_mapping, on=CONTENT_ID, how=LEFT)
            seasonal_content_df = seasonal_content_df[~seasonal_content_df[HOMEPAGE_ID].isnull()].reset_index(drop=True)
            seasonal_content_df = seasonal_content_df.explode(column=HOMEPAGE_ID)
            seasonal_content_df = SeasonalModel.filter_by_dedicated_homepage_id(seasonal_content_df, user_label,
                                                                                filter_on_ubd=False)
            seasonal_content_df = seasonal_content_df.sort_values(by=[HOMEPAGE_ID, VIEW_COUNT], ascending=[True, False])
        else:
            seasonal_content_df = pd.merge(seasonal_content_df, content_homepage_id_mapping, on=CONTENT_ID, how=LEFT)
            seasonal_content_df = seasonal_content_df[~seasonal_content_df[HOMEPAGE_ID].isnull()].reset_index(drop=True)
            seasonal_content_df = seasonal_content_df.explode(column=HOMEPAGE_ID)
            seasonal_content_df = SeasonalModel.filter_by_dedicated_homepage_id(seasonal_content_df, user_label,
                                                                                filter_on_ubd=False)
            seasonal_content_df = seasonal_content_df.sort_values(by=VIEW_COUNT, ascending=False)
        return seasonal_content_df

    @staticmethod
    def get_default_seasonal_content_all_content(graph):
        list_of_key_prefix = []
        homepage_id_wise = False
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Seasonal Content based on All Content - DEFAULT MODE")
        start = time.time()
        for user_label in list_user_label:
            Logging.info("Start generating seasonal content based on All Content for " + user_label + " users - DEFAULT MODE")
            output_df = SeasonalModelDefault.default_seasonal_algorithm(graph, user_label, homepage_id_wise)
            Logging.info("Success Preparing Default Seasonal Dataframe Output - DEFAULT MODE")
            key_prefix = SERVICE_NAME + ":" + DEFAULT_SEASONAL_MODULE_NAME + ":" + user_label + ":" + ALL_CONTENT_BASED
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        default_seasonal_pay_tv_all_content, default_seasonal_no_pay_tv_all_content = list_output_df

        Logging.info("Preparing the output to JSON Schema")
        # Convert dataframe output to json output
        default_seasonal_pay_tv_all_content_dict = SeasonalModelDefault.get_dict_format_output(
            default_seasonal_pay_tv_all_content, list_of_key_prefix[0], homepage_id_wise)
        default_seasonal_no_pay_tv_all_content_dict = SeasonalModelDefault.get_dict_format_output(
            default_seasonal_no_pay_tv_all_content, list_of_key_prefix[1], homepage_id_wise)
        end = time.time()
        duration = end - start
        Logging.info("Duration for Get Default Seasonal Content based on All Content is " + str(duration) + " seconds")
        return default_seasonal_pay_tv_all_content_dict, default_seasonal_no_pay_tv_all_content_dict

    @staticmethod
    def get_default_seasonal_content_homepage_id(graph):
        list_of_key_prefix = []
        homepage_id_wise = True
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Seasonal Content based on Homepage_id - DEFAULT MODE")
        start = time.time()
        for user_label in list_user_label:
            Logging.info("Start generating seasonal content based on Homepage_id for " + user_label + " users - DEFAULT MODE")
            output_df = SeasonalModelDefault.default_seasonal_algorithm(graph, user_label, homepage_id_wise)
            Logging.info("Success Preparing Default Seasonal Dataframe Output - DEFAULT MODE")
            key_prefix = SERVICE_NAME + ":" + DEFAULT_SEASONAL_MODULE_NAME + ":" + user_label + ":" + HOMEPAGE_ID_BASED
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        default_seasonal_pay_tv_homepage_id, default_seasonal_no_pay_tv_homepage_id = list_output_df

        Logging.info("Preparing the output to JSON Schema")
        # Convert dataframe output to json output
        default_seasonal_pay_tv_homepage_id_dict = SeasonalModelDefault.get_dict_format_output(default_seasonal_pay_tv_homepage_id,
                                                                                               list_of_key_prefix[0],
                                                                                               homepage_id_wise)
        default_seasonal_no_pay_tv_homepage_id_dict = SeasonalModelDefault.get_dict_format_output(default_seasonal_no_pay_tv_homepage_id,
                                                                                                  list_of_key_prefix[1],
                                                                                                  homepage_id_wise)

        end = time.time()
        duration = end - start
        Logging.info("Duration for Get Default Seasonal Content based on Homepage_id is " + str(duration) + " seconds")
        return default_seasonal_pay_tv_homepage_id_dict, default_seasonal_no_pay_tv_homepage_id_dict
