from typing import ClassVar

import ast
from operator import itemgetter
import pandas as pd
from graphdb import GraphDb, GraphDbConnection
from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler

from offline_results.common.constants import USER_LABEL, USER_CONTENT_RELATIONSHIP_LABEL, CONTENT_ID, \
    CREATED_ON, VIEW_HISTORY, VIEW_COUNT, UBD_WEEKOFYEAR, DURATION, UBD_CREATED_ON, UBD_DAYOFWEEK, CONTENT_CREATED_ON, \
    SCALED_DURATION, SCALED_VIEW_COUNT, INNER, TRENDING_SCORE


class TrendingModel:
    def __init__(
            self,
            connection_class: GraphDbConnection
    ):
        self.graph = GraphDb.from_connection(connection_class)

    @classmethod
    def from_connection_uri(
            cls,
            connection_uri: str
    ) -> ClassVar:
        """Create new object based on connection uri
        :param connection_uri: string connection uri
        :return: object class
        """
        return cls(GraphDbConnection.from_uri(connection_uri))

    @classmethod
    def from_connection_class(
            cls,
            connection_class: GraphDbConnection
    ) -> ClassVar:
        """Define new class based on object connection
        :param connection_class: object connection class
        :return: object class
        """
        return cls(connection_class)

    @staticmethod
    def get_history_date(history):
        history_data = history[1][VIEW_HISTORY]
        list_history_data = ast.literal_eval(history_data)
        list_history_data = sorted(list_history_data, key=itemgetter(CREATED_ON), reverse=True)
        history_date = pd.DataFrame(list_history_data).rename(columns={CREATED_ON: UBD_CREATED_ON})
        return history_date

    @staticmethod
    def fetch_historical_data(graph):
        views_query = graph.custom_query(f'''
                g.V().hasLabel('{USER_LABEL}').outE().hasLabel('{USER_CONTENT_RELATIONSHIP_LABEL}').inV().path().by(elementMap())
                ''', payload={
            USER_LABEL: USER_LABEL,
            USER_CONTENT_RELATIONSHIP_LABEL: USER_CONTENT_RELATIONSHIP_LABEL
        })
        history_content_df = DataFrame()
        for view in views_query:
            for history in view:
                content_object_data = history[2][CONTENT_ID]
                content_created_date = history[2][CREATED_ON]
                history_date_data = TrendingModel.get_history_date(history)
                history_date_data[CONTENT_ID] = content_object_data
                history_date_data[CONTENT_CREATED_ON] = content_created_date
                history_content_df = pd.concat([history_content_df, history_date_data], axis=0)
        history_content_df = history_content_df.reset_index(drop=True)
        history_content_df[CONTENT_CREATED_ON] = pd.to_datetime(history_content_df[CONTENT_CREATED_ON], dayfirst=True)
        history_content_df[UBD_CREATED_ON] = pd.to_datetime(history_content_df[UBD_CREATED_ON], dayfirst=True)
        history_content_df = history_content_df.sort_values(by=[UBD_CREATED_ON, CONTENT_CREATED_ON],
                                                            ascending=False).reset_index(drop=True)
        history_content_df[UBD_WEEKOFYEAR] = history_content_df[UBD_CREATED_ON].dt.isocalendar().week
        history_content_df[UBD_DAYOFWEEK] = history_content_df[UBD_CREATED_ON].dt.dayofweek
        tv_channel_list = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 17, 22, 38, 46, 48, 52, 55, 60, 62, 64, 2374, 2440,
                           2449, 2770, 2755, 2830, 2833, 2901, 2909, 2933, 6209, 6537, 6545, 6549, 6553, 7426, 7432,
                           7414, 19881]
        history_content_df_vod = history_content_df[~history_content_df[CONTENT_ID].isin(tv_channel_list)]\
            .reset_index(drop=True)
        return history_content_df, history_content_df_vod

    @staticmethod
    def get_latest_week_df(history_content_df):
        get_latest_week_number = history_content_df[UBD_WEEKOFYEAR].unique()[0]
        latest_week_df = history_content_df[history_content_df[UBD_WEEKOFYEAR] == get_latest_week_number]
        total_dayofweek = len(latest_week_df[UBD_DAYOFWEEK].unique())
        if total_dayofweek < 4:
            two_latest_week_number = history_content_df[UBD_WEEKOFYEAR].unique()[0:2].tolist()
            latest_week_df = history_content_df[history_content_df[UBD_WEEKOFYEAR].isin(two_latest_week_number)]
        else:
            return latest_week_df

        return latest_week_df

    @staticmethod
    def get_trending_content(graph):
        history_content_df, history_content_df_vod = TrendingModel.fetch_historical_data(graph)
        latest_week_df = TrendingModel.get_latest_week_df(history_content_df_vod)

        view_count_on_latest_week_df = latest_week_df.groupby([CONTENT_ID]).size().to_frame(VIEW_COUNT)
        total_duration_content = latest_week_df.groupby([CONTENT_ID]).sum()
        total_duration_content = total_duration_content[[DURATION]]
        final_recommendation_df = pd.merge(total_duration_content, view_count_on_latest_week_df,
                                           left_index=True, right_index=True).sort_values(by=[DURATION, VIEW_COUNT],
                                                                                          ascending=False)
        scaled_features_df = MinMaxScaler().fit_transform(final_recommendation_df.values)
        scaled_features_df = pd.DataFrame(scaled_features_df, index=final_recommendation_df.index,
                                          columns=final_recommendation_df.columns).rename(
            columns={DURATION: SCALED_DURATION, VIEW_COUNT: SCALED_VIEW_COUNT}
        )
        final_recommendation_df = final_recommendation_df.merge(scaled_features_df, how=INNER,
                                                                left_index=True, right_index=True)
        final_recommendation_df[TRENDING_SCORE] = (final_recommendation_df[SCALED_DURATION]*0.7) + \
                                                  (final_recommendation_df[SCALED_VIEW_COUNT]*0.3)
        final_recommendation_df = final_recommendation_df.sort_values(by=TRENDING_SCORE, ascending=False)
        list_trending_content = final_recommendation_df.head(10).index.tolist()
        return history_content_df, latest_week_df, final_recommendation_df, list_trending_content