from pandas import DataFrame, merge, concat, to_timedelta, Series
import numpy as np

from offline_results.repository import RedisConnection
from datetime import datetime
from graphdb import GraphDb
from graphdb.connection import GraphDbConnection
from offline_results.common.constants import PREF_FEATURES, CSV_EXTENSION, \
    CUSTOMER_ID, VIDEO_ID1, DURATION, USER_LABEL, PAYTVPROVIDER_ID, \
    RELATIONSHIP, MESSAGE, RATING, ATTRIBUTE1, \
    CATEGORY, SUBCATEGORY, ACTORS, DIRECTORS, TAGS, IS_PAYTV, EDGE
from offline_results.common.config import BUCKET_NAME, USER_MAPPING_KEY, \
    DURATION_CUTOFF, HAS_PAYTV_PROVIDER
from offline_results.similarity.user_profile.config import CLUSTER_NODE_LABEL, \
    NEW_USER_CLUSTER_RELATIONSHIP_LABEL
from offline_results.common.read_write_from_s3 import ConnectS3


class ClusterUpdater:

    def __init__(self, ubd_new, graph_connection):
        """Assumption service that include this module already create connection
        then passing that connection here, no need create new connection
        :param graph_connection: object graphdb connection
        :param redis_connection: connection uri redis
        """
        self.data = ubd_new
        # self.redis = redis_connection
        self.graph = GraphDb.from_connection(graph_connection)

    def find_cutoff_users(
            self,
    ) -> DataFrame:
        """
        Find customer_ids that have duration above
        specified threshold after grouping.
        :return: Dataframe object pandas
        """
        data = self.data
        data[CUSTOMER_ID] = data[CUSTOMER_ID].astype(str)
        data = data[data[CUSTOMER_ID] != "0"]
        data = data[data.created_on > datetime.now() - to_timedelta("20 days")]
        data = data.groupby([CUSTOMER_ID, VIDEO_ID1])[DURATION].sum().reset_index()
        data = data[data[DURATION] > DURATION_CUTOFF].reset_index(drop=True)
        data = data.drop(columns=[DURATION])
        data = merge(self.data, data, on=[CUSTOMER_ID, VIDEO_ID1], how='inner')

        return data

    def fetch_updated_pref(
            self,
            bucket_name=None,
            object_name=None,
            resource=None
    ) -> DataFrame:
        """
        Fetches preferences from S3 and merges in
        one file for the considered users.
        :param bucket_name: s3 bucket name
        :param object_name: s3 key
        :param resource: boto connection resource
        :return: all preference dataframe
        for considered users
        """
        data = self.find_cutoff_users()
        updated_pref = DataFrame()
        updated_pref[CUSTOMER_ID] = data[CUSTOMER_ID].unique()

        for feature in PREF_FEATURES:
            s3_data = ConnectS3().read_csv_from_s3(bucket_name=BUCKET_NAME,
                                                   object_name=USER_MAPPING_KEY + feature + CSV_EXTENSION,
                                                   resource=resource)
            print("Successfully read {}.csv from S3...".format(feature))
            s3_data[CUSTOMER_ID] = s3_data[CUSTOMER_ID].astype(str)
            s3_data = s3_data[s3_data[CUSTOMER_ID].isin(data[CUSTOMER_ID].tolist())]

            print("Merging updated {} preference dataframe into a single dataframe...".format(feature))
            updated_pref = updated_pref.merge(s3_data, on=CUSTOMER_ID, how='outer')

        return updated_pref

    def filter_user_preferences(
            self,
            data: DataFrame
    ) -> DataFrame:
        """
        Function to assign column names to rows
        where user preference is 1.
        :param data: Dataframe object pandas
        :return: cleansed dataframe object pandas
        """
        data = data.loc[:, (data != 0).any(axis=0)]
        cleansed_data = data.where(data != 1, data.columns.to_series(), axis=1)

        return cleansed_data

    def fetch_existing_cluster(
            self,
            user_list: list
    ) -> DataFrame:
        """
        Query the graph for cluster labels of the
        considered users and save them in DataFrame
        :return: Dataframe of customer_id and
        current assigned cluster label
        """
        existing_user_clusters = DataFrame()
        for user in user_list:
            response = self.graph.custom_query(
                query=f'''g.V().has('{USER_LABEL}','{CUSTOMER_ID}', '{user}').
                outE('{NEW_USER_CLUSTER_RELATIONSHIP_LABEL}').
                project('{EDGE}','{CUSTOMER_ID}','{CLUSTER_NODE_LABEL}').
                by(label).
                by(outV().values('{CUSTOMER_ID}')).
                by(inV().union(
                values('{CLUSTER_NODE_LABEL}'),
                values('is_pay_tv')).fold())'''
            )
            if len(response) > 0:
                row_to_add = response[0][0]
                existing_user_clusters = concat([existing_user_clusters, DataFrame([row_to_add])], ignore_index=True)

        return existing_user_clusters

    def find_paytv_users(
            self,
            data: DataFrame
    ):
        """
        Separate users paytv and non paytv wise
        :param data: list of customer_ids
        """
        user_list = data[CUSTOMER_ID].tolist()
        users = DataFrame()
        for user in user_list:
            response = self.graph.custom_query(
                query=f'''g.V().has('{USER_LABEL}','{CUSTOMER_ID}', '{user}').
                outE('{HAS_PAYTV_PROVIDER}').project('{EDGE}','{CUSTOMER_ID}','{PAYTVPROVIDER_ID}').
                by(label).
                by(outV().values('{CUSTOMER_ID}')).
                by(inV().values('{PAYTVPROVIDER_ID}'))'''
            )
            if len(response) > 0:
                row_to_add = response[0][0]
                users = concat([users, DataFrame([row_to_add])], ignore_index=True)

        if not users.empty:
            users = users.drop(columns=[EDGE])
            data = data.merge(users, on=CUSTOMER_ID, how='left')
        else:
            data[PAYTVPROVIDER_ID] = np.nan

        return data

    def fetch_centroids(
            self
    ):
        """
        Fetch kmeans centroid from the graph
        """
        response = self.graph.custom_query(
            query=f'''g.V().hasLabel('centroid').has('{CLUSTER_NODE_LABEL}').
            path().by(elementMap())'''
        )
        tmp = {}
        centroids = DataFrame()
        for index, value in enumerate(response):
            for i, j in enumerate(value):
                tmp[CLUSTER_NODE_LABEL] = j[0][CLUSTER_NODE_LABEL]
                tmp[IS_PAYTV] = j[0][IS_PAYTV]
                tmp[RATING] = j[0][RATING]
                tmp[ATTRIBUTE1] = j[0][ATTRIBUTE1]
                tmp[CATEGORY] = j[0][CATEGORY]
                tmp[SUBCATEGORY] = j[0][SUBCATEGORY]
                tmp[ACTORS] = j[0][ACTORS]
                tmp[DIRECTORS] = j[0][DIRECTORS]
                tmp[TAGS] = j[0][TAGS]

                centroids = concat([centroids, DataFrame([tmp])])

        return centroids.reset_index(drop=True)

    def prepare_data(
            self,
            user: DataFrame,
    ) -> DataFrame:
        """
        Function to prepare user data for further processing
        :param user: User dataframe object pandas
        :return: Dataframe object pandas
        """
        user_data = DataFrame()
        user_data[CUSTOMER_ID] = user[CUSTOMER_ID]
        user_data = user_data.set_index(CUSTOMER_ID)
        for feature in PREF_FEATURES:
            tmp = user.set_index(CUSTOMER_ID)
            tmp = tmp.loc[:, ~tmp.columns.str.endswith(("_nan", "_"))]
            mask = Series(tmp.columns).str.contains(feature)
            tmp = tmp[tmp.columns[mask]]
            tmp[feature] = tmp.iloc[:, 0:].apply(
                lambda i: [''.join(str(i))
                           for i in i.where(i.str.startswith(feature)).
                           dropna().tolist() if i != 0],
                axis=1)
            tmp = tmp.drop(columns=tmp.columns[tmp.columns.str.contains(feature + "_")].tolist())
            user_data = user_data.merge(tmp, left_index=True, right_index=True)

        return user_data.reset_index().rename(columns={'index': CUSTOMER_ID})

    def fetch_user_centroid_rel(
            self,
            user: DataFrame,
            centroid: DataFrame
    ) -> DataFrame:
        """
        Function to compare each column of user with
        each column of centroid df
        :param user: Dataframe object pandas
        :param centroid: Dataframe object pandas
        :return: Dataframe object pandas
        """
        tmp_df = DataFrame()
        tmp_dict = {}

        for index1, row1 in user.iterrows():
            for index2, row2 in centroid.iterrows():
                tmp_dict[CUSTOMER_ID] = row1[CUSTOMER_ID]
                tmp_dict[CLUSTER_NODE_LABEL] = row2[CLUSTER_NODE_LABEL]
                tmp_dict[IS_PAYTV] = row2[IS_PAYTV]
                for i in PREF_FEATURES:
                    intersec = set(row1[i]).intersection(set(row2[2:][i]))
                    tmp_dict[i] = len(intersec)
                    print("working on customer_id:{} index:{}".format(row1[CUSTOMER_ID], index2))

                tmp_df = concat([tmp_df, DataFrame([tmp_dict])])

        tmp_df['sum'] = tmp_df.iloc[:, 3:].sum(1)

        return tmp_df.reset_index(drop=True)

    def fetch_new_clusters(
            self,
            data: DataFrame
    ) -> DataFrame:
        """
        Function to assign new cluster labels to users
        :param data: Dataframe object pandas
        :return: Dataframe object pandas
        """
        new_cluster_df = DataFrame()
        for j in data[CUSTOMER_ID].unique():
            sample_df = data[data[CUSTOMER_ID] == j]
            max_sum = max(sample_df['sum'])
            filtered_tmp_df = sample_df[sample_df['sum'] == max_sum]
            sample_cluster_id = filtered_tmp_df.sample()
            cluster_id = sample_cluster_id.reset_index(drop=True)
            cluster_id = cluster_id.iloc[:, 0:3]
            new_cluster_df = concat([new_cluster_df, cluster_id])

        return new_cluster_df.reset_index(drop=True)

    def assign_new_clusters(
            self,
            user: DataFrame,
            centroids: DataFrame
    ):
        """
        Function to assign new cluster labels to users
        :param user: User dataframe object pandas
        :param centroids: centroid dataframe object
        :return: Dataframe object pandas
        """
        for feature in PREF_FEATURES:
            centroids[feature] = [x.split(',') for x in centroids[feature]]

        centroids_true = centroids[centroids[IS_PAYTV] == 'True'].reset_index(drop=True)
        centroids_false = centroids[centroids[IS_PAYTV] == 'False'].reset_index(drop=True)

        no_paytv_user = user[user[PAYTVPROVIDER_ID].isna()]
        no_paytv_user = no_paytv_user.drop(columns=PAYTVPROVIDER_ID).reset_index(drop=True)
        no_paytv_user = no_paytv_user.loc[:, (no_paytv_user != 0).any(axis=0)]
        no_paytv_user = self.prepare_data(user=no_paytv_user)

        print("Fetching user centroid relation for no_paytv_users...")
        no_paytv_user_centroid_rel = self.fetch_user_centroid_rel(user=no_paytv_user,
                                                                  centroid=centroids_false)

        print("Fetching new cluster labels for no_paytv_users...")
        new_no_paytv_user_cluster = self.fetch_new_clusters(data=no_paytv_user_centroid_rel)

        if not user[~user[PAYTVPROVIDER_ID].isna()].empty:
            paytv_user = user[~user[PAYTVPROVIDER_ID].isna()]
            paytv_user = paytv_user.drop(columns=PAYTVPROVIDER_ID).reset_index(drop=True)
            paytv_user = paytv_user.loc[:, (paytv_user != 0).any(axis=0)]
            paytv_user = self.prepare_data(user=paytv_user)
            print("Fetching user centroid relation for paytv_users...")
            paytv_user_centroid_rel = self.fetch_user_centroid_rel(user=paytv_user,
                                                                   centroid=centroids_true)
            print("Fetching new cluster labels for paytv_users...")
            new_paytv_user_cluster = self.fetch_new_clusters(data=paytv_user_centroid_rel)

            new_user_clusters = concat([new_paytv_user_cluster, new_no_paytv_user_cluster],
                                       ignore_index=True)
            return new_user_clusters

        return new_no_paytv_user_cluster

    def save_cluster_relation_s3(
            self,
            cluster_data: DataFrame,
            bucket_name=None,
            object_name=None,
            resource=None
    ):
        """
        Function to save updated user-cluster relations on S3
        :param cluster_data: user-cluster relation dataframe
        :param bucket_name: s3 bucket name
        :param object_name: s3 file path
        :param resource: boto connection resource
        """
        print("Saving updated user-cluster relations to S3...")
        ConnectS3().write_csv_to_s3(bucket_name=BUCKET_NAME,
                                    object_name=USER_MAPPING_KEY + 'updated_user_clusters.csv',
                                    df_to_upload=cluster_data,
                                    resource=resource)

    def drop_existing_clusters(
            self,
            existing_clusters_df: DataFrame
    ):
        """
        Drop existing user-cluster relation from graph
        :param existing_clusters_df: Dataframe object pandas
        """
        list_user = existing_clusters_df[CUSTOMER_ID].unique().tolist()
        for user in list_user:
            print("Dropping {} for customer:{}".format(existing_clusters_df[0, EDGE], user))
            self.graph.custom_query(
                query=f'''g.V().hasLabel('{USER_LABEL}').
                has('{CUSTOMER_ID}', '{user}').outE('{NEW_USER_CLUSTER_RELATIONSHIP_LABEL}').drop()'''
            )

    def plot_new_cluster_relations(
            self,
            users_cluster: DataFrame
    ):
        """
        Function to plot new cluster relations in graphdb
        :param users_cluster: Dataframe object pandas
        """
        undumped_rel = DataFrame()
        for index in range(len(users_cluster)):
            print("Dumping relation {} of {}".format(index + 1, len(users_cluster)))

            try:
                source_node = self.graph.custom_query(
                    query=f'''g.V().hasLabel('{USER_LABEL}').
                                has('{CUSTOMER_ID}', '{users_cluster.loc[index, CUSTOMER_ID]}')'''
                )[0][0].id

                destination_node = self.graph.custom_query(
                    query=f'''g.V().hasLabel('{CLUSTER_NODE_LABEL}').
                                has('{CLUSTER_NODE_LABEL}', {users_cluster.loc[index, CLUSTER_NODE_LABEL]}).
                                has('is_pay_tv', '{users_cluster.loc[index, IS_PAYTV]}')'''
                )[0][0].id

                print("Creating relationship between customer_id:{} and {}:{}".
                      format(users_cluster.loc[index, CUSTOMER_ID], CLUSTER_NODE_LABEL,
                             users_cluster.loc[index, CLUSTER_NODE_LABEL])
                      )

                self.graph.custom_query(
                    query=f'''g.V('{source_node}').as('v').V('{destination_node}').
                                coalesce(__.inE('{NEW_USER_CLUSTER_RELATIONSHIP_LABEL}').where(outV().as('v')),
                                addE('{NEW_USER_CLUSTER_RELATIONSHIP_LABEL}').from('v'))'''
                )

            except Exception as e:
                print("Unable to create HAS_MINIBATCH_CLUSTER relation between customer_id:{} and {}:{}".
                      format(users_cluster.loc[index, CUSTOMER_ID],
                             CLUSTER_NODE_LABEL, users_cluster.loc[index, CLUSTER_NODE_LABEL])
                      )
                error_dumping = DataFrame.from_dict({CUSTOMER_ID: [users_cluster.loc[index, CUSTOMER_ID]],
                                                     RELATIONSHIP: NEW_USER_CLUSTER_RELATIONSHIP_LABEL,
                                                     CLUSTER_NODE_LABEL: [users_cluster.loc[index, CLUSTER_NODE_LABEL]],
                                                     MESSAGE: e})
                undumped_rel = concat([undumped_rel, error_dumping], axis=0, ignore_index=True)

        undumped_rel.to_csv("undumped_" + CLUSTER_NODE_LABEL + "_relations.csv", index=False)

    def controller(
            self,
            bucket_name=None,
            object_name=None,
            resource=None
    ):

        data = self.fetch_updated_pref(bucket_name=bucket_name,
                                       object_name=object_name,
                                       resource=resource)
        data = self.filter_user_preferences(data)
        existing_clusters = self.fetch_existing_cluster(data[CUSTOMER_ID].tolist())
        data = self.find_paytv_users(data)
        centroid_data = self.fetch_centroids()
        new_user_clusters = self.assign_new_clusters(user=data, centroids=centroid_data)
        self.save_cluster_relation_s3(cluster_data=new_user_clusters,
                                      bucket_name=bucket_name,
                                      object_name=object_name,
                                      resource=resource)
        self.drop_existing_clusters(existing_clusters)
        self.plot_new_cluster_relations(users_cluster=new_user_clusters)
